@@ -42,7 +42,7 @@
 
 #include <openobex/obex.h>
 
-#include "logging.h"
+#include "log.h"
 #include "obex.h"
 #include "obex-priv.h"
 #include "server.h"
@@ -56,6 +56,18 @@
 #define DEFAULT_RX_MTU 32767
 #define DEFAULT_TX_MTU 32767
 
+/* Challenge request */
+#define NONCE_TAG 0x00
+#define OPTIONS_TAG 0x01 /* Optional */
+#define REALM_TAG 0x02 /* Optional */
+
+#define NONCE_LEN 16
+
+/* Challenge response */
+#define DIGEST_TAG 0x00
+#define USER_ID_TAG 0x01 /* Optional */
+#define DIGEST_NONCE_TAG 0x02 /* Optional */
+
 /* Connection ID */
 static uint32_t cid = 0x0000;
 
@@ -67,6 +79,144 @@ typedef struct {
 	uint16_t mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
+struct auth_header {
+	uint8_t tag;
+	uint8_t len;
+	uint8_t val[0];
+} __attribute__ ((packed));
+
+static struct {
+	int evt;
+	const char *name;
+} obex_event[] = {
+	/* Progress has been made */
+	{ OBEX_EV_PROGRESS,	"PROGRESS"	},
+	/* An incoming request is about to come */
+	{ OBEX_EV_REQHINT,	"REQHINT"	},
+	/* An incoming request has arrived */
+	{ OBEX_EV_REQ,		"REQ"		},
+	/* Request has finished */
+	{ OBEX_EV_REQDONE,	"REQDONE"	},
+	/* Link has been disconnected */
+	{ OBEX_EV_LINKERR,	"LINKERR"	},
+	/* Malformed data encountered */
+	{ OBEX_EV_PARSEERR,	"PARSEERR"	},
+	/* Connection accepted */
+	{ OBEX_EV_ACCEPTHINT,	"ACCEPTHINT"	},
+	/* Request was aborted */
+	{ OBEX_EV_ABORT,	"ABORT"		},
+	/* Need to feed more data when sending a stream */
+	{ OBEX_EV_STREAMEMPTY,	"STREAMEMPTY"	},
+	/* Time to pick up data when receiving a stream */
+	{ OBEX_EV_STREAMAVAIL,	"STREAMAVAIL"	},
+	/* Unexpected data, not fatal */
+	{ OBEX_EV_UNEXPECTED,	"UNEXPECTED"	},
+	/* First packet of an incoming request has been parsed */
+	{ OBEX_EV_REQCHECK,	"REQCHECK"	},
+	{ 0xFF,			NULL		},
+};
+
+/* Possible commands */
+static struct {
+	int cmd;
+	const char *name;
+} obex_command[] = {
+	{ OBEX_CMD_CONNECT,	"CONNECT"	},
+	{ OBEX_CMD_DISCONNECT,	"DISCONNECT"	},
+	{ OBEX_CMD_PUT,		"PUT"		},
+	{ OBEX_CMD_GET,		"GET"		},
+	{ OBEX_CMD_SETPATH,	"SETPATH"	},
+	{ OBEX_CMD_SESSION,	"SESSION"	},
+	{ OBEX_CMD_ABORT,	"ABORT"		},
+	{ OBEX_FINAL,		"FINAL"		},
+	{ 0xFF,			NULL		},
+};
+
+/* Possible Response */
+static struct {
+	int rsp;
+	const char *name;
+} obex_response[] = {
+	{ OBEX_RSP_CONTINUE,			"CONTINUE"		},
+	{ OBEX_RSP_SWITCH_PRO,			"SWITCH_PRO"		},
+	{ OBEX_RSP_SUCCESS,			"SUCCESS"		},
+	{ OBEX_RSP_CREATED,			"CREATED"		},
+	{ OBEX_RSP_ACCEPTED,			"ACCEPTED"		},
+	{ OBEX_RSP_NON_AUTHORITATIVE,		"NON_AUTHORITATIVE"	},
+	{ OBEX_RSP_NO_CONTENT,			"NO_CONTENT"		},
+	{ OBEX_RSP_RESET_CONTENT,		"RESET_CONTENT"		},
+	{ OBEX_RSP_PARTIAL_CONTENT,		"PARTIAL_CONTENT"	},
+	{ OBEX_RSP_MULTIPLE_CHOICES,		"MULTIPLE_CHOICES"	},
+	{ OBEX_RSP_MOVED_PERMANENTLY,		"MOVED_PERMANENTLY"	},
+	{ OBEX_RSP_MOVED_TEMPORARILY,		"MOVED_TEMPORARILY"	},
+	{ OBEX_RSP_SEE_OTHER,			"SEE_OTHER"		},
+	{ OBEX_RSP_NOT_MODIFIED,		"NOT_MODIFIED"		},
+	{ OBEX_RSP_USE_PROXY,			"USE_PROXY"		},
+	{ OBEX_RSP_BAD_REQUEST,			"BAD_REQUEST"		},
+	{ OBEX_RSP_UNAUTHORIZED,		"UNAUTHORIZED"		},
+	{ OBEX_RSP_PAYMENT_REQUIRED,		"PAYMENT_REQUIRED"	},
+	{ OBEX_RSP_FORBIDDEN,			"FORBIDDEN"		},
+	{ OBEX_RSP_NOT_FOUND,			"NOT_FOUND"		},
+	{ OBEX_RSP_METHOD_NOT_ALLOWED,		"METHOD_NOT_ALLOWED"	},
+	{ OBEX_RSP_NOT_ACCEPTABLE,		"NOT_ACCEPTABLE"	},
+	{ OBEX_RSP_PROXY_AUTH_REQUIRED,		"PROXY_AUTH_REQUIRED"	},
+	{ OBEX_RSP_REQUEST_TIME_OUT,		"REQUEST_TIME_OUT"	},
+	{ OBEX_RSP_CONFLICT,			"CONFLICT"		},
+	{ OBEX_RSP_GONE,			"GONE"			},
+	{ OBEX_RSP_LENGTH_REQUIRED,		"LENGTH_REQUIRED"	},
+	{ OBEX_RSP_PRECONDITION_FAILED,		"PRECONDITION_FAILED"	},
+	{ OBEX_RSP_REQ_ENTITY_TOO_LARGE,	"REQ_ENTITY_TOO_LARGE"	},
+	{ OBEX_RSP_REQ_URL_TOO_LARGE,		"REQ_URL_TOO_LARGE"	},
+	{ OBEX_RSP_UNSUPPORTED_MEDIA_TYPE,	"UNSUPPORTED_MEDIA_TYPE"},
+	{ OBEX_RSP_INTERNAL_SERVER_ERROR,	"INTERNAL_SERVER_ERROR"	},
+	{ OBEX_RSP_NOT_IMPLEMENTED,		"NOT_IMPLEMENTED"	},
+	{ OBEX_RSP_BAD_GATEWAY,			"BAD_GATEWAY"		},
+	{ OBEX_RSP_SERVICE_UNAVAILABLE,		"SERVICE_UNAVAILABLE"	},
+	{ OBEX_RSP_GATEWAY_TIMEOUT,		"GATEWAY_TIMEOUT"	},
+	{ OBEX_RSP_VERSION_NOT_SUPPORTED,	"VERSION_NOT_SUPPORTED"	},
+	{ OBEX_RSP_DATABASE_FULL,		"DATABASE_FULL"		},
+	{ OBEX_RSP_DATABASE_LOCKED,		"DATABASE_LOCKED"	},
+	{ 0xFF,					NULL			},
+};
+
+static void print_event(int evt, int cmd, int rsp)
+{
+	const char *evtstr = NULL, *cmdstr = NULL, *rspstr = NULL;
+	int i;
+	static int lastevt, lastcmd;
+
+	if (evt < 0)
+		evt = lastevt;
+	else
+		lastevt = evt;
+
+	if (cmd < 0)
+		cmd = lastcmd;
+	else
+		lastcmd = cmd;
+
+	for (i = 0; obex_event[i].evt != 0xFF; i++) {
+		if (obex_event[i].evt != evt)
+			continue;
+		evtstr = obex_event[i].name;
+	}
+
+	for (i = 0; obex_command[i].cmd != 0xFF; i++) {
+		if (obex_command[i].cmd != cmd)
+			continue;
+		cmdstr = obex_command[i].name;
+	}
+
+	for (i = 0; obex_response[i].rsp != 0xFF; i++) {
+		if (obex_response[i].rsp != rsp)
+			continue;
+		rspstr = obex_response[i].name;
+	}
+
+	obex_debug("%s(0x%x), %s(0x%x), %s(0x%x)", evtstr, evt, cmdstr, cmd,
+								rspstr, rsp);
+}
+
 static void os_set_response(obex_object_t *obj, int err)
 {
 	uint8_t rsp;
@@ -108,17 +258,33 @@ static void os_set_response(obex_object_t *obj, int err)
 		lastrsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
 	}
 
+	print_event(-1, -1, rsp);
+
 	OBEX_ObjectSetRsp(obj, rsp, lastrsp);
 }
 
+static void os_session_mark_aborted(struct obex_session *os)
+{
+	/* the session was already cancelled/aborted or size in unknown */
+	if (os->aborted || os->size == OBJECT_SIZE_UNKNOWN)
+		return;
+
+	os->aborted = (os->size != os->offset);
+}
+
 static void os_reset_session(struct obex_session *os)
 {
+	os_session_mark_aborted(os);
+	if (os->service && os->service->reset)
+		os->service->reset(os, os->service_data);
+
 	if (os->object) {
 		os->driver->set_io_watch(os->object, NULL, NULL);
 		os->driver->close(os->object);
 		os->object = NULL;
 		os->obj = NULL;
-		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path)
+		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path &&
+				os->driver->remove)
 			os->driver->remove(os->path);
 	}
 
@@ -147,16 +313,6 @@ static void os_reset_session(struct obex_session *os)
 	os->finished = 0;
 }
 
-static void os_session_mark_aborted(struct obex_session *os)
-{
-	/* the session was alredy cancelled/aborted */
-	if (os->aborted)
-		return;
-
-	os->aborted = os->size == OBJECT_SIZE_UNKNOWN ? FALSE :
-							os->size != os->offset;
-}
-
 static void obex_session_free(struct obex_session *os)
 {
 	sessions = g_slist_remove(sessions, os);
@@ -216,6 +372,53 @@ static time_t parse_iso8610(const char *val, int size)
 	return time;
 }
 
+static uint8_t *extract_nonce(const uint8_t *buffer, unsigned int hlen)
+{
+	struct auth_header *hdr;
+	uint8_t *nonce = NULL;
+	uint32_t len = 0;
+
+	while (len < hlen) {
+		hdr = (void *) buffer + len;
+
+		switch (hdr->tag) {
+		case NONCE_TAG:
+			if (hdr->len != NONCE_LEN)
+				return NULL;
+
+			nonce = hdr->val;
+			break;
+		}
+
+		len += hdr->len + sizeof(struct auth_header);
+	}
+
+	return nonce;
+}
+
+static uint8_t *challenge_response(const uint8_t *nonce)
+{
+	GChecksum *md5;
+	uint8_t *result;
+	size_t size;
+
+	result = g_new0(uint8_t, NONCE_LEN);
+
+	md5 = g_checksum_new(G_CHECKSUM_MD5);
+	if (md5 == NULL)
+		return result;
+
+	g_checksum_update(md5, nonce, NONCE_LEN);
+	g_checksum_update(md5, (uint8_t *) ":BlueZ", 6);
+
+	size = NONCE_LEN;
+	g_checksum_get_digest(md5, result, &size);
+
+	g_checksum_free(md5);
+
+	return result;
+}
+
 static void cmd_connect(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
@@ -225,26 +428,26 @@ static void cmd_connect(struct obex_session *os,
 	unsigned int hlen, newsize;
 	uint16_t mtu;
 	uint8_t hi;
-	const uint8_t *target = NULL, *who = NULL;
+	const uint8_t *target = NULL, *who = NULL, *nonce = NULL;
 	unsigned int target_size = 0, who_size = 0;
 	int err;
 
 	if (OBEX_ObjectGetNonHdrData(obj, &buffer) != sizeof(*nonhdr)) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		debug("Invalid OBEX CONNECT packet");
+		DBG("Invalid OBEX CONNECT packet");
 		return;
 	}
 
 	nonhdr = (obex_connect_hdr_t *) buffer;
 	mtu = g_ntohs(nonhdr->mtu);
-	debug("Version: 0x%02x. Flags: 0x%02x  OBEX packet length: %d",
+	DBG("Version: 0x%02x. Flags: 0x%02x  OBEX packet length: %d",
 			nonhdr->version, nonhdr->flags, mtu);
 	/* Leave space for headers */
 	newsize = mtu - 200;
 
 	os->tx_mtu = newsize;
 
-	debug("Resizing stream chunks to %d", newsize);
+	DBG("Resizing stream chunks to %d", newsize);
 
 	/* connection id will be used to track the sessions, even for OPP */
 	os->cid = ++cid;
@@ -259,6 +462,15 @@ static void cmd_connect(struct obex_session *os,
 			target = hd.bs;
 			target_size = hlen;
 			break;
+		case OBEX_HDR_AUTHCHAL:
+			if (nonce) {
+				DBG("Ignoring multiple challenge headers");
+				break;
+			}
+
+			nonce = extract_nonce(hd.bs, hlen);
+			DBG("AUTH CHALLENGE REQUEST");
+			break;
 		}
 	}
 
@@ -272,7 +484,7 @@ static void cmd_connect(struct obex_session *os,
 		return;
 	}
 
-	debug("Selected driver: %s", os->service->name);
+	DBG("Selected driver: %s", os->service->name);
 
 	if (!os->service->connect) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -291,6 +503,21 @@ static void cmd_connect(struct obex_session *os,
 				OBEX_FL_FIT_ONE_PACKET);
 	}
 
+	if (err == 0 && nonce) {
+		uint8_t challenge[18];
+		struct auth_header *hdr = (struct auth_header *) challenge;
+		uint8_t *response = challenge_response(nonce);
+
+		hdr->tag = DIGEST_TAG;
+		hdr->len = NONCE_LEN;
+		memcpy(hdr->val, response, NONCE_LEN);
+
+		g_free(response);
+
+		hd.bs = challenge;
+		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_AUTHRESP, hd, 18, 0);
+	}
+
 	os_set_response(obj, err);
 }
 
@@ -331,6 +558,10 @@ static int obex_read_stream(struct obex_session *os, obex_t *obex,
 	int32_t len = 0;
 	const uint8_t *buffer;
 
+	DBG("name=%s type=%s rx_mtu=%d file=%p",
+		os->name ? os->name : "", os->type ? os->type : "",
+		os->rx_mtu, os->object);
+
 	if (os->aborted)
 		return -EPERM;
 
@@ -357,7 +588,7 @@ static int obex_read_stream(struct obex_session *os, obex_t *obex,
 	memcpy(os->buf + os->pending, buffer, size);
 	os->pending += size;
 	if (os->object == NULL) {
-		debug("Stored %u bytes into temporary buffer", os->pending);
+		DBG("Stored %u bytes into temporary buffer", os->pending);
 		return 0;
 	}
 
@@ -393,7 +624,7 @@ static int obex_write_stream(struct obex_session *os,
 	unsigned int flags;
 	uint8_t hi;
 
-	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
+	DBG("name=%s type=%s tx_mtu=%d file=%p",
 		os->name ? os->name : "", os->type ? os->type : "",
 		os->tx_mtu, os->object);
 
@@ -467,28 +698,10 @@ static gboolean handle_async_io(void *object, int flags, int err,
 		ret = obex_read_stream(os, os->obex, os->obj);
 
 proceed:
-	switch (ret) {
-	case -EINVAL:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		break;
-	case -EPERM:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
-					OBEX_RSP_FORBIDDEN);
-		break;
-	case -ENOENT:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_NOT_FOUND,
-							OBEX_RSP_NOT_FOUND);
-		break;
-	default:
-		if (ret < 0)
-			OBEX_ObjectSetRsp(os->obj,
-					OBEX_RSP_INTERNAL_SERVER_ERROR,
-					OBEX_RSP_INTERNAL_SERVER_ERROR);
-		break;
-	}
-
-	OBEX_ResumeRequest(os->obex);
+	if (ret < 0)
+		OBEX_CancelRequest(os->obex, TRUE);
+	else
+		OBEX_ResumeRequest(os->obex);
 
 	return FALSE;
 }
@@ -516,7 +729,7 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 		switch (hi) {
 		case OBEX_HDR_NAME:
 			if (os->name) {
-				debug("Ignoring multiple name headers");
+				DBG("Ignoring multiple name headers");
 				break;
 			}
 
@@ -525,11 +738,11 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 
 			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
-			debug("OBEX_HDR_NAME: %s", os->name);
+			DBG("OBEX_HDR_NAME: %s", os->name);
 			break;
 		case OBEX_HDR_TYPE:
 			if (os->type) {
-				debug("Ignoring multiple type headers");
+				DBG("Ignoring multiple type headers");
 				break;
 			}
 
@@ -541,16 +754,18 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 				break;
 
 			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
-				debug("Invalid type header: %s", hd.bs);
+				DBG("Invalid type header: %s", hd.bs);
 				break;
 			}
 
 			/* FIXME: x-obex/folder-listing - type is mandatory */
 
 			os->type = g_strndup((const char *) hd.bs, hlen);
-			debug("OBEX_HDR_TYPE: %s", os->type);
+			DBG("OBEX_HDR_TYPE: %s", os->type);
 			os->driver = obex_mime_type_driver_find(
-						os->service->target, os->type,
+						os->service->target,
+						os->service->target_size,
+						os->type,
 						os->service->who,
 						os->service->who_size);
 			break;
@@ -559,6 +774,7 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 
 	if (os->type == NULL)
 		os->driver = obex_mime_type_driver_find(os->service->target,
+							os->service->target_size,
 							NULL,
 							os->service->who,
 							os->service->who_size);
@@ -630,7 +846,7 @@ static void cmd_setpath(struct obex_session *os,
 			continue;
 
 		if (os->name) {
-			debug("Ignoring multiple name headers");
+			DBG("Ignoring multiple name headers");
 			break;
 		}
 
@@ -644,7 +860,7 @@ static void cmd_setpath(struct obex_session *os,
 		os->name = g_convert((const char *) hd.bs, hlen,
 				"UTF8", "UTF16BE", NULL, NULL, NULL);
 
-		debug("Set path name: %s", os->name);
+		DBG("Set path name: %s", os->name);
 		break;
 	}
 
@@ -656,7 +872,7 @@ int obex_get_stream_start(struct obex_session *os, const char *filename)
 {
 	int err;
 	void *object;
-	size_t size;
+	size_t size = OBJECT_SIZE_UNKNOWN;
 
 	object = os->driver->open(filename, O_RDONLY, 0, os->service_data,
 								&size, &err);
@@ -697,7 +913,7 @@ int obex_put_stream_start(struct obex_session *os, const char *filename)
 	os->path = g_strdup(filename);
 
 	if (!os->buf) {
-		debug("PUT request checked, no buffered data");
+		DBG("PUT request checked, no buffered data");
 		return 0;
 	}
 
@@ -731,7 +947,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		switch (hi) {
 		case OBEX_HDR_NAME:
 			if (os->name) {
-				debug("Ignoring multiple name headers");
+				DBG("Ignoring multiple name headers");
 				break;
 			}
 
@@ -740,12 +956,12 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
-			debug("OBEX_HDR_NAME: %s", os->name);
+			DBG("OBEX_HDR_NAME: %s", os->name);
 			break;
 
 		case OBEX_HDR_TYPE:
 			if (os->type) {
-				debug("Ignoring multiple type headers");
+				DBG("Ignoring multiple type headers");
 				break;
 			}
 
@@ -757,14 +973,16 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 				break;
 
 			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
-				debug("Invalid type header: %s", hd.bs);
+				DBG("Invalid type header: %s", hd.bs);
 				break;
 			}
 
 			os->type = g_strndup((const char *) hd.bs, hlen);
-			debug("OBEX_HDR_TYPE: %s", os->type);
+			DBG("OBEX_HDR_TYPE: %s", os->type);
 			os->driver = obex_mime_type_driver_find(
-						os->service->target, os->type,
+						os->service->target,
+						os->service->target_size,
+						os->type,
 						os->service->who,
 						os->service->who_size);
 			break;
@@ -776,7 +994,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 		case OBEX_HDR_LENGTH:
 			os->size = hd.bq4;
-			debug("OBEX_HDR_LENGTH: %d", os->size);
+			DBG("OBEX_HDR_LENGTH: %d", os->size);
 			break;
 		case OBEX_HDR_TIME:
 			os->time = parse_iso8610((const char *) hd.bs, hlen);
@@ -788,6 +1006,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 	if (os->type == NULL)
 		os->driver = obex_mime_type_driver_find(os->service->target,
+							os->service->target_size,
 							NULL,
 							os->service->who,
 							os->service->who_size);
@@ -819,7 +1038,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		os->driver->set_io_watch(os->object, handle_async_io, os);
 		return TRUE;
 	default:
-		debug("Unhandled chkput error: %d", ret);
+		DBG("Unhandled chkput error: %d", ret);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
 				OBEX_RSP_INTERNAL_SERVER_ERROR);
 		return FALSE;
@@ -827,7 +1046,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	}
 
 	if (os->size == OBJECT_SIZE_DELETE || os->size == OBJECT_SIZE_UNKNOWN) {
-		debug("Got a PUT without a Length");
+		DBG("Got a PUT without a Length");
 		goto done;
 	}
 
@@ -844,10 +1063,6 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
-	} else if (!os->service->put) {
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
-				OBEX_RSP_NOT_IMPLEMENTED);
-		return;
 	}
 
 	g_return_if_fail(chk_cid(obex, obj, os->cid));
@@ -857,29 +1072,34 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			return;
 	}
 
+	if (!os->service->put) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+				OBEX_RSP_NOT_IMPLEMENTED);
+		return;
+	}
+
 	err = os->service->put(os, os->service_data);
 	if (err < 0)
 		os_set_response(obj, err);
 }
 
-static void obex_event(obex_t *obex, obex_object_t *obj, int mode,
+static void obex_event_cb(obex_t *obex, obex_object_t *obj, int mode,
 					int evt, int cmd, int rsp)
 {
 	struct obex_session *os;
+	int err;
 
-	obex_debug(evt, cmd, rsp);
+	print_event(evt, cmd, rsp);
 
 	os = OBEX_GetUserData(obex);
 
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
-		if (os->service->progress)
+		if (os->service && os->service->progress)
 			os->service->progress(os, os->service_data);
 		break;
 	case OBEX_EV_ABORT:
 		os->aborted = TRUE;
-		if (os->service->reset)
-			os->service->reset(os, os->service_data);
 		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
@@ -894,9 +1114,6 @@ static void obex_event(obex_t *obex, obex_object_t *obj, int mode,
 		case OBEX_CMD_GET:
 		case OBEX_CMD_SETPATH:
 		default:
-			os_session_mark_aborted(os);
-			if (os->service->reset)
-				os->service->reset(os, os->service_data);
 			os_reset_session(os);
 			break;
 		}
@@ -947,49 +1164,32 @@ static void obex_event(obex_t *obex, obex_object_t *obj, int mode,
 			cmd_put(os, obex, obj);
 			break;
 		default:
-			debug("Unknown request: 0x%X", cmd);
+			DBG("Unknown request: 0x%X", cmd);
 			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 						OBEX_RSP_NOT_IMPLEMENTED);
 			break;
 		}
 		break;
 	case OBEX_EV_STREAMAVAIL:
-		switch (obex_read_stream(os, obex, obj)) {
-		case 0:
-			break;
-		case -EPERM:
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-			break;
-		case -EAGAIN:
+		err = obex_read_stream(os, obex, obj);
+		if (err == -EAGAIN) {
 			OBEX_SuspendRequest(obex, obj);
 			os->obj = obj;
 			os->driver->set_io_watch(os->object, handle_async_io,
 									os);
-			break;
-		default:
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_INTERNAL_SERVER_ERROR,
-				OBEX_RSP_INTERNAL_SERVER_ERROR);
-			break;
-		}
+		} else if (err < 0)
+			os_set_response(obj, err);
 
 		break;
 	case OBEX_EV_STREAMEMPTY:
-		switch (obex_write_stream(os, obex, obj)) {
-		case -EPERM:
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN,
-							OBEX_RSP_FORBIDDEN);
-			break;
-		case -EAGAIN:
+		err = obex_write_stream(os, obex, obj);
+		if (err == -EAGAIN) {
 			OBEX_SuspendRequest(obex, obj);
 			os->obj = obj;
 			os->driver->set_io_watch(os->object, handle_async_io,
 									os);
-			break;
-		default:
-			break;
-		}
+		} else if (err < 0)
+			os_set_response(obj, err);
 
 		break;
 	case OBEX_EV_LINKERR:
@@ -1000,7 +1200,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, int mode,
 		break;
 
 	default:
-		debug("Unknown evt %d", evt);
+		DBG("Unknown evt %d", evt);
 		break;
 	}
 }
@@ -1010,6 +1210,8 @@ static void obex_handle_destroy(void *user_data)
 	struct obex_session *os;
 	obex_t *obex = user_data;
 
+	DBG("");
+
 	os = OBEX_GetUserData(obex);
 
 	if (os->service && os->service->disconnect)
@@ -1057,7 +1259,7 @@ int obex_session_start(GIOChannel *io, uint16_t tx_mtu, uint16_t rx_mtu,
 	os->tx_mtu = tx_mtu != 0 ? tx_mtu : DEFAULT_TX_MTU;
 	os->size = OBJECT_SIZE_DELETE;
 
-	obex = OBEX_Init(OBEX_TRANS_FD, obex_event, 0);
+	obex = OBEX_Init(OBEX_TRANS_FD, obex_event_cb, 0);
 	if (!obex) {
 		obex_session_free(os);
 		return -EIO;
@@ -1096,7 +1298,7 @@ void obex_set_name(struct obex_session *os, const char *name)
 {
 	g_free(os->name);
 	os->name = g_strdup(name);
-	debug("Name changed: %s", os->name);
+	DBG("Name changed: %s", os->name);
 }
 
 ssize_t obex_get_size(struct obex_session *os)
@@ -1189,13 +1391,16 @@ int obex_aparam_write(struct obex_session *os,
 			OBEX_HDR_APPARAM, hd, size, 0);
 }
 
-int memcmp0(const void *a, const void *b, size_t n)
+int memncmp0(const void *a, size_t na, const void *b, size_t nb)
 {
+	if (na != nb)
+		return na - nb;
+
 	if (a == NULL)
 		return -(a != b);
 
 	if (b == NULL)
 		return a != b;
 
-	return memcmp(a, b, n);
+	return memcmp(a, b, na);
 }
